/**
 * 
 * controller缓存、$scope生命周期、观察者模式
 * ------------------------------------------------------------------------------------------------------------------------
 * 
 * 定义：
 * 观察者模式(有时又被称为发布-订阅Subscribe>模式、模型-视图View>模式、源-收听者Listener>模式或从属者模式)是软件设计模式的一种。
 * 在此种模式中，一个目标物件管理所有相依于它的观察者物件，并且在它本身的状态改变时主动发出通知。
 * 这通常透过呼叫各观察者所提供的方法来实现。此种模式通常被用来实现事件处理系统。
 * 
 * 需求：
 * 1. 在angularJS中，如果某state启用了缓存（cache:true），那么除第一次进入该state之外，后续进入都不再重新加载。
 * 2. 但是，该state（A）中的部分数据可能受到后续state（B）的影响。例如，状态A中有一个数据，状态B可以修改它。从状态B转入状态A后，状态A的UI应该做对应的改变。
 * 
 * 解决方案：
 * 1. 在状态A中注册响应的监听器（一个功能函数），
 * 2. 当状态B修改了可能影响到状态A的数据（事件）的时候，状态B发送广播消息通知状态A，
 * 3. 状态A的监听器接监听到广播，根据广播参数做出对应响应。这实际上是有广播协调者调用A的监听器完成的。
 * 4. ※ 如果离开状态A，状态A应该主动反注册已经注册的监听器。
 * 
 * 注册监听器与反注册监听器的时机：
 * 1. 注册：$scope并没有“$create”事件，所以，就在controller中任意位置写吧。^_^
 * 2. 反注册：$scope提供了一个“$destory”事件，通过监听“$destory”事件（广播）来反注册。
 * 
 * 
 * 
 * -------------------------------------------------------------------------------------------------------------------------------------------------------------
 * 补充：
 * 实验发现：
 * 1. 对于cache为true的状态，除了当前状态的缓存被清除的情况外，即便是跳转到其它状态或者重新进入，否则“$destory”事件不会发生。
 *    这也是为什么通过监听“$destory”事件来反注册广播监听器的原因。
 * 2. 对于cache为false的状态，只要离开当前状态，就会触发“$destory”事件。
 * 
 * 
 * 
 * 参考：
 * 1.$rootScope.Scope
 * https://docs.angularjs.org/api/ng/type/$rootScope.Scope
 * Events
 *  $destroy
 *      Broadcasted when a scope and its children are being destroyed.
 *      Note that, in AngularJS, there is also a $destroy jQuery event, which can be used to clean up DOM bindings before an element is removed from the DOM.
 *  Type:
 *      broadcast
 *  Target:
 *      scope being destroyed
 * 
 * 2. $rootScope.Scope
 * https://docs.angularjs.org/api/ng/type/$rootScope.Scope
 * $on(name, listener);
 * $emit(name, args);
 * $broadcast(name, args);
 * 
 * 3. angularjs的$on、$emit、$broadcast（该机制无法实现兄弟controller间的广播，尽管没有什么用，但是可以帮助理解事件广播原理。）
 * http://www.cnblogs.com/yuzhongwusan/p/4938816.html
 * 
 */




/**
 * 该部分服务属于公共服务。
 * 该部分服务是一些自定的基础服务。
 * 主要是一些通用的广播（消息）事件调用。
 * 本服务实际上是使用了“观察者”模式来实现的。
 */

// UI服务 
angular.module('starter.services')
    .factory('broadcastService', [function () {

        // 广播池（数组）。元素是： {name:string, listeners:function[]}
        // name：广播名称
        // listeners: name对应的监听器池（数组）
        // listener: 监听器（功能函数），参数是data；如果以有序方式发送广播，且某listener返回true，则不再通知其后面的监听器。
        var broadcasts = [];


        /**
         *
         * 注册广播监听器
         * name: 广播名称
         * listener: 监听器（处理函数）
         * allowRepeated: 在listener已经存在于监听器数组中的情况下是否允许再次插入
         * 
         */
        var register = function (name, listener, allowRepeated) {
            if (!name) {
                return;
            }
            // init
            if (!broadcasts) {
                broadcasts = [];
            }
            // get broadcast
            var broadcast;
            for (var i = 0; i < broadcasts.length; i++) {
                var tmpBroadcast = broadcasts[i];
                if (tmpBroadcast) {
                    if (tmpBroadcast.name === name) {
                        broadcast = tmpBroadcast;
                        break;
                    }
                }
            }
            if (!broadcast) {
                // init broadcast
                broadcast = {
                    name: name,
                    listeners: []
                };
                broadcasts.push(broadcast);
            }
            // insert listener into listeners
            var listeners = broadcast.listeners;
            if (allowRepeated) {
                listeners.push(listener);
            } else {
                var found = false;
                for (var i = 0; i < listeners; i++) {
                    if (listeners[i] == listener) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    listeners.push(listener);
                }
            }
        };

        /**
         *
         * 反注册广播监听器
         * name: 广播名称
         * listener: 监听器（处理函数）
         * 
         */
        var unregister = function (name, listener) {
            if (!name) {
                return;
            }
            if (broadcasts) {
                for (var i = 0; i < broadcasts.length; i++) {
                    var broadcast = broadcasts[i];
                    if (broadcast.name == name) {
                        var listeners = broadcast.listeners;
                        if (listeners) {
                            // 移除监听器    可以移除多个相同的
                            for (var j = listeners.length - 1; j >= 0; j--) {
                                if (listeners[j] == listener) {
                                    listeners.splice(j, 1);
                                }
                            }
                        }
                        // 对于监听器个数为0的，可以考虑清楚对应的广播
                        if (!listeners || listeners.length == 0) {
                            broadcasts.splice(i, 1);
                        }
                        // 假设程序员的素质很高，能够保证broadcast中插入的broadcast的name属性是唯一的。
                        break;
                    }
                }
            }
        };

        /**
         * 反注册所有的监听器；如果指定的name不为空，则只反注册该name对应的监听器；如果name为false或者null，或者是undefined，则反注册所有name的监听器。 
         */
        var unregisterAll = function (name) {
            if (name == null || name == undefined) {
                broadcasts = null;
            } else {
                for (var i = 0; i < broadcasts.length; i++) {
                    var broadcast = broadcasts[i];
                    if (broadcast.name == name) {
                        broadcasts.splice(i, 1);
                        // 假设程序员的素质很高，能够保证broadcast中插入的broadcast的name属性是唯一的。
                        break;
                    }
                }
            }
        };

        /**
         *
         * 发送广播
         * name: 广播名称
         * data: 监听器（处理函数）接收的参数对象
         * ordered: 是否是有序广播
         * 
         * -------------------------------------------------------------------------------------------------------------------------------------------------------------
         * 1. 广播分为有序广播和无序广播。
         *    其中，有序广播是同步执行的，只用当前的监听器处理完广播，后面的广播才可以继续执行；并且，当前监听器返回true的时候可以阻止通知后面的监听器。
         *    无序广播是以Promise对象的方式异步执行的，没有先后顺序。
         * 2. 对广播事件的处理需要通过监听器来实现，即注册和反注册（监听系统的$destory事件）；
         */
        var send = function (name, data, ordered) {
            if (!name) {
                return;
            }
            //
            if (broadcasts) {
                for (var i = 0; i < broadcasts.length; i++) {
                    var broadcast = broadcasts[i];
                    if (broadcast.name == name) {
                        var listeners = broadcast.listeners;
                        if (listeners) {
                            for (var j = listeners.length - 1; j >= 0; j--) {
                                var listener = listeners[j];
                                if (listener) {
                                    if (ordered) {
                                        var prevented = listener(data);
                                        if (prevented == true) {
                                            break; // 如果返回了true，表示不再需要继续传递下去
                                        }
                                    } else {
                                        // 注：JavaScript Promise迷你书（中文版） http://liubin.org/promises-book/
                                        new Promise(
                                            function (resolve, reject) {
                                                resolve(data);
                                            }
                                        ).then(listener, null);
                                    }
                                }
                            }
                        }
                        // 假设程序员的素质很高，能够保证broadcast中插入的broadcast的name属性是唯一的。
                        break;
                    }
                }
            }
        };




        // --------------------------------------------------------------------------------------------------------------------------------------
        return {
            register: register,
            unregister: unregister,
            unregisterAll: unregisterAll,
            send: send,
        };

    }]);